import argparse
from typing import List

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_files_glob import (
    arg_files_glob,
    arg_file_glob_compile_files,
    arg_files_print,
)
from ul_py_tool.utils.step import Stepper


class CmdFmt(Cmd):
    py_file_lists: List[List[str]]
    py_file_ignore_lists: List[List[str]]
    print_files: int

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--print-files",
            dest="print_files",
            type=int,
            required=False,
            default=10,
        )
        parser.add_argument(
            "--py-files",
            dest="py_file_lists",
            nargs="+",
            type=arg_files_glob(),
            required=False,
            default=[
                arg_files_glob(ignore_absent=True)("*.py"),
                arg_files_glob(ignore_absent=True)("**/*.py"),
            ],
        )
        parser.add_argument(
            "--py-files-ignore",
            dest="py_file_ignore_lists",
            nargs="+",
            type=arg_files_glob(ignore_absent=True),
            required=False,
            default=[],
        )

    def run(self) -> None:
        py_files, ignored_files = arg_file_glob_compile_files(
            self.py_file_lists,
            self.py_file_ignore_lists,
        )
        arg_files_print(
            self.print_files,
            py_files,
            ignored_files=ignored_files,
            name="python files",
        )

        stepper = Stepper()

        # stp.step_cmd('ISORT', [
        #     'python', '-m', 'isort',
        #     '--profile=pycharm',
        #     '--order-by-type',
        #     '--diff',
        #     '--trailing-comma',
        #     '--honor-noqa',
        #     '--color',
        #     # '--force-single-line-imports',
        #     '--force-alphabetical-sort',
        #     '--lines-after-imports=2',
        #     '--lines-between-types=1',
        #     '--settings-path', os.path.join(THIS_LIB_PATH, 'conf', 'isort.cfg'),
        #     # '--show-config',
        #     # '--no-sections',
        #     '--line-length', '180',
        #     # '--sections=FUTURE,STDLIB,FIRSTPARTY,THIRDPARTY,LOCALFOLDER',
        #     # '--check-only',
        #     *py_files,
        # ], chk=self._check_isort, print_std=False)

        with stepper.step("BLACK") as stp:
            stp.run_cmd(["black", *py_files])



import os

import yaml

from ul_py_tool.utils.docker_compose import DockerComposeFile
from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.colors import FG_GREEN, NC
from ul_py_tool.utils.run_command import run_command
from ul_py_tool.utils.step import Stepper


class CmdReleaseDcfDev(Cmd):
    """
    Uses local files and values to generate test DCF
    """
    def run(self) -> None:
        stepper = Stepper()
        cwd = os.getcwd()

        with stepper.step(f'{FG_GREEN}release{NC}'):
            run_command(
                [
                    'werf render --config=werf.yaml --values=.helm/dev-values.yaml  --values=.helm/charts.yaml --secret-values=.helm/dev-secret-values.yaml --loose-giterminism=True > kubernetes-objects.yaml',
                ],
                cwd=cwd,
            )
            with open(os.path.join(cwd, "kubernetes-objects.yaml"), "r") as fr:
                dcf = DockerComposeFile.from_kubernetes(yaml.load_all(fr, yaml.FullLoader))
            final_file = dcf.to_json()
            with open("docker-compose.release.yaml", "w") as fw:
                yaml.dump(final_file, fw)



import argparse
from typing import List, Tuple, Dict

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_files_glob import arg_files_glob, arg_file_glob_compile_files, arg_files_print
from ul_py_tool.utils.colors import FG_YELLOW, NC
from ul_py_tool.utils.write_stdout import write_stdout


class CmdStats(Cmd):
    print_files: int
    file_lists: List[List[str]]
    file_ignore_lists: List[List[str]]

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        finder = arg_files_glob(ignore_absent=True)
        parser.add_argument('--print-files', dest='print_files', type=int, required=False, default=10)
        parser.add_argument('--files', dest='file_lists', nargs='+', type=finder, required=False, default=[
            finder('*'),
            finder('**/*'),
        ])
        parser.add_argument('--ignore', dest='file_ignore_lists', nargs='+', type=finder, required=False, default=[])

    def run(self) -> None:
        py_files, ignored_files = arg_file_glob_compile_files(self.file_lists, self.file_ignore_lists)
        arg_files_print(self.print_files, py_files, ignored_files=ignored_files, name='files to count', ide_col_suf=False)

        write_stdout('\n')

        stats_by_type: Dict[str, List[Tuple[str, int, int]]] = dict()
        max_pf = 0
        max_pf_t = 0
        for pf in py_files:
            max_pf = max(len(pf), max_pf)
            try:
                with open(pf, 'rt') as f:
                    this_file_count_not_empty_lines = 0
                    this_file_count_lines = 0
                    for line in f.readlines():
                        this_file_count_lines += 1
                        if len(line.strip()) > 0:
                            this_file_count_not_empty_lines += 1
                    pft = pf.split('/')[-1].split('.')[-1]
                    max_pf_t = max(len(pft), max_pf_t)
                    if pft not in stats_by_type:
                        stats_by_type[pft] = []
                    stats_by_type[pft].append((pf, this_file_count_lines, this_file_count_not_empty_lines))
            except UnicodeDecodeError:
                continue

        count_lines = 0
        count_not_empty_lines = 0
        for type, stats in stats_by_type.items():
            t_count_lines = 0
            t_count_not_empty_lines = 0
            for _file_path, count, non_empty_count in stats:
                count_not_empty_lines += non_empty_count
                count_lines += count
                t_count_not_empty_lines += non_empty_count
                t_count_lines += count
                # write_stdout(f'{arg_file_info(file_path, ide_col_suf=True, rel_path=True): <{max_pf + 10:}} = {FG_YELLOW}{non_empty_count: >4}{NC} / {count: <4} lines')
            write_stdout(f'{len(stats): >4} *.{type: <{max_pf_t}} files has {FG_YELLOW}{t_count_not_empty_lines}{NC} / {t_count_lines} lines')
        write_stdout(f'{"=" * 80}\nsummary: {len(py_files)} files has {FG_YELLOW}{count_not_empty_lines}{NC} / {count_lines} lines\n')


def test_example() -> None:
    assert 1 == 1



from typing import List

from ul_py_tool.utils.write_stdout import write_stdout


def input_lines(name: str) -> List[str]:
    logs = []
    write_stdout(f'Please add {name} (leave empty if don\'t)')
    while True:
        text = input(': ')
        if not text:
            break
        logs.append(text)
    return logs



import argparse
import os
from typing import List

import yaml
from deepdiff import DeepDiff  # type: ignore

from ul_py_tool.commands.conf import HELM_ERROR__NOT_FOUND

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_str2yaml import arg_str2yaml
from ul_py_tool.utils.aseembly import AssemblyFile, AssemblyTarget
from ul_py_tool.utils.colors import FG_GREEN, FG_RED, NC
from ul_py_tool.utils.run_command import run_command
from ul_py_tool.utils.step import Stepper


class CmdTestSecrets(Cmd):
    release_environment: str
    assembly_file: AssemblyFile
    assembly_targets: List[AssemblyTarget] = []

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--release-environment', dest='release_environment', type=str, required=True)
        parser.add_argument('--assembly-file', dest='assembly_file', type=arg_str2yaml, required=True, default='assembly.yaml')

    def run(self) -> None:
        stepper = Stepper()
        cwd = os.getcwd()

        with stepper.step(f'{FG_GREEN}check if target exists{NC}'):
            self.assembly_targets = [at for at in self.assembly_file.targets.values() if at.environment == self.release_environment]

        with stepper.step(f'{FG_GREEN}check secrets{NC}') as stp:
            for component_name in self.assembly_file.components.keys():
                for at in self.assembly_targets:
                    values_dir = os.path.join(cwd, 'secrets', component_name, f'{at.cluster_name}:{at.environment}')
                    if not os.path.exists(os.path.join(values_dir, 'secret-values.yaml')):
                        stp.add_error(f'{FG_RED}for {component_name} secrets not found in {values_dir}{NC}')
                    if not os.path.exists(os.path.join(values_dir, 'values.yaml')):
                        stp.add_error(f'{FG_RED}for {component_name} deploy values not found in {values_dir}{NC}')

        with stepper.step(f'{FG_GREEN}init required repositories{NC}'):
            for component in self.assembly_file.components.values():
                os.makedirs(os.path.join(cwd, component.directory), exist_ok=False)
                run_command([f'git clone  --depth 1 --branch {component.tag} {component.repository} .'], cwd=os.path.join(cwd, component.directory))
                run_command(['git submodule sync --recursive && git submodule update --init --recursive'], cwd=os.path.join(cwd, component.directory))

        with stepper.step(f'{FG_GREEN}check values and secrets data{NC}') as stp:
            ''' format of DeepDiff output looks like:
            {
              'dictionary_item_added': [root['broker'], root['db'], root['dockerconfig'], root['systemconfig'], root['pii_keys']],
              'dictionary_item_removed': [root['endpoints']]
            }
            or just {} if no diff in keys
            where dictionary_item_added - keys that are not presented  in first dict(yaml)(dev-values/dev-secret-values.yaml)
            and dictionary_item_removed - keys that are presented in first dict(yaml)(dev-values/dev-secret-values.yaml) but missing in second (target)
            '''
            for component_name, component_config in self.assembly_file.components.items():
                for at in self.assembly_targets:
                    values_dir = os.path.join(cwd, 'secrets', component_name, f'{at.cluster_name}:{at.environment}')

                    with open(os.path.join(cwd, component_config.directory, '.helm/dev-values.yaml'), 'r') as f:
                        dev_v = yaml.safe_load(f)

                    with open(os.path.join(values_dir, "values.yaml"), 'r') as f:
                        target_v = yaml.safe_load(f)
                    ddiff = DeepDiff(dev_v, target_v, ignore_order=True)
                    ddiff.pop('values_changed', None)
                    if ddiff:
                        stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in values:{NC}\n {ddiff}')

                    with open(os.path.join(cwd, component_config.directory, '.helm/dev-secret-values.yaml'), 'r') as f:
                        dev_secret_v = yaml.safe_load(f)

                    with open(os.path.join(values_dir, "secret-values.yaml"), 'r') as f:
                        target_secret_v = yaml.safe_load(f)
                    ddiff = DeepDiff(dev_secret_v, target_secret_v, ignore_order=True)
                    ddiff.pop('values_changed', None)
                    if ddiff:
                        stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in secret values:{NC}\n {ddiff}')

                    # test if helm chart exists to avoid "oops we deployed but it crashed on last step because bad package and no new features
                    helm_chart = run_command(
                        [
                            f'helm search repo {component_config.helm_repo_name}/{component_config.helm_chart_name} --version {component_config.helm_chart_version}',
                        ],
                    ).stdout
                    if helm_chart == HELM_ERROR__NOT_FOUND:
                        raise NotImplementedError(
                            f'{component_config.helm_chart_name} chart '
                            f'at local or remote repo {component_config.helm_repo_name} '
                            f'of version {component_config.helm_chart_version} was not found',
                        )



from typing import Dict

from pydantic import BaseModel


class AssemblyComponent(BaseModel):
    repository: str
    directory: str
    tag: str
    force_disable: bool
    helm_repo_name: str
    helm_chart_name: str
    helm_chart_version: str


class AssemblyTarget(BaseModel):
    environment: str
    cluster_name: str
    kubeconfig_file: str
    encryption_key_name: str
    recreate: bool = False


class AssemblyFile(BaseModel):
    targets: Dict[str, AssemblyTarget]
    components: Dict[str, AssemblyComponent]



import argparse
import os
import shutil
from typing import Optional, Dict
import time
from kubernetes import client, config  # type: ignore
import yaml
from deepdiff import DeepDiff  # type: ignore

from ul_py_tool.commands.conf import HELM_ERROR__NOT_FOUND, KUEBERNETES_SLEEP_TIME

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.utils.arg_str2yaml import arg_str2yaml
from ul_py_tool.utils.aseembly import AssemblyFile, AssemblyTarget
from ul_py_tool.utils.colors import FG_GREEN, FG_RED, NC
from ul_py_tool.utils.run_command import run_command
from ul_py_tool.utils.step import Stepper
from ul_py_tool.utils.write_stdout import write_stdout


class CmdRelease(Cmd):
    release_target: str
    assembly_file: AssemblyFile
    assembly_target: Optional[AssemblyTarget] = None

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--release-target', dest='release_target', type=str, required=True)
        parser.add_argument('--assembly-file', dest='assembly_file', type=arg_str2yaml, required=True, default='assembly.yaml')

    def run(self) -> None:
        stepper = Stepper()
        cwd = os.getcwd()

        with stepper.step(f'{FG_GREEN}check if target exists{NC}'):
            if self.release_target not in self.assembly_file.targets.keys():
                raise argparse.ArgumentTypeError(f'{FG_RED}Target not found in assembly file{NC}')
            self.assembly_target = self.assembly_file.targets[self.release_target]

        with stepper.step(f'{FG_GREEN}check secrets{NC}') as stp:
            for component_name in self.assembly_file.components.keys():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')
                if not os.path.exists(os.path.join(values_dir, 'secret-values.yaml')):
                    stp.add_error(f'{FG_RED}for {component_name} secrets not found in {values_dir}{NC}')
                if not os.path.exists(os.path.join(values_dir, 'values.yaml')):
                    stp.add_error(f'{FG_RED}for {component_name} deploy values not found in {values_dir}{NC}')

        with stepper.step(f'{FG_GREEN}init required repositories{NC}'):
            for component in self.assembly_file.components.values():
                os.makedirs(os.path.join(cwd, component.directory), exist_ok=False)
                run_command([f'git clone  --depth 1 --branch {component.tag} {component.repository} .'], cwd=os.path.join(cwd, component.directory))
                run_command(['git submodule sync --recursive && git submodule update --init --recursive'], cwd=os.path.join(cwd, component.directory))

        with stepper.step(f'{FG_GREEN}check values and secrets data{NC}') as stp:
            ''' format of DeepDiff output looks like:
            {
              'dictionary_item_added': [root['broker'], root['db'], root['dockerconfig'], root['systemconfig'], root['pii_keys']],
              'dictionary_item_removed': [root['endpoints']]
            }
            or just {} if no diff in keys
            where dictionary_item_added - keys that are not presented  in first dict(yaml)(dev-values/dev-secret-values.yaml)
            and dictionary_item_removed - keys that are presented in first dict(yaml)(dev-values/dev-secret-values.yaml) but missing in second (target)
            '''
            for component_name, component_config in self.assembly_file.components.items():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')

                with open(os.path.join(cwd, component_config.directory, '.helm/dev-values.yaml'), 'r') as f:
                    dev_v = yaml.safe_load(f)

                with open(os.path.join(values_dir, "values.yaml"), 'r') as f:
                    target_v = yaml.safe_load(f)
                ddiff = DeepDiff(dev_v, target_v, ignore_order=True)
                ddiff.pop('values_changed', None)
                if ddiff:
                    stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in values:{NC}\n {ddiff}')

                with open(os.path.join(cwd, component_config.directory, '.helm/dev-secret-values.yaml'), 'r') as f:
                    dev_secret_v = yaml.safe_load(f)

                with open(os.path.join(values_dir, "secret-values.yaml"), 'r') as f:
                    target_secret_v = yaml.safe_load(f)
                ddiff = DeepDiff(dev_secret_v, target_secret_v, ignore_order=True)
                ddiff.pop('values_changed', None)
                if ddiff:
                    stp.add_error(f'{FG_RED}for {component_name} deploy values have missing or extra keys in secret values:{NC}\n {ddiff}')

                # test if helm chart exists to avoid "oops we deployed but it crashed on last step because bad package and no new features
                helm_chart = run_command(
                    [
                        f'helm search repo {component_config.helm_repo_name}/{component_config.helm_chart_name} --version {component_config.helm_chart_version}',
                    ],
                ).stdout
                if helm_chart == HELM_ERROR__NOT_FOUND:
                    raise NotImplementedError(
                        f'{component_config.helm_chart_name} chart '
                        f'at local or remote repo {component_config.helm_repo_name} '
                        f'of version {component_config.helm_chart_version} was not found',
                    )

        with stepper.step(f'{FG_GREEN}release{NC}'):
            # to store releases that completed
            releases: Dict[str, str] = {}
            kubeconfig_file_path = os.path.join(cwd, self.assembly_target.kubeconfig_file)
            config.load_kube_config(config_file=kubeconfig_file_path)
            kubernetes_client = client.AppsV1Api()
            for component_name, component_config in self.assembly_file.components.items():
                values_dir = os.path.join(cwd, 'secrets', component_name, f'{self.assembly_target.cluster_name}:{self.assembly_target.environment}')
                component_directory = os.path.join(cwd, component_config.directory)
                release_name = f'{component_name}-{self.assembly_target.environment}'

                # get list of all up-to-date releases
                list_deployments = kubernetes_client.list_namespaced_deployment(
                    namespace=release_name,
                ).items

                # collect all unique labels (after 1.1.26 ul-py-tool they are compulsory, bot now not)
                set_charts_versions = set()
                set_service_versions = set()
                set_deployment_names = set()
                for deployment in list_deployments:
                    set_charts_versions.add(deployment.metadata.labels.get('release/chart-version', None))
                    set_service_versions.add(deployment.metadata.labels.get('release/component-version', None))
                    set_deployment_names.add(deployment.metadata.name)

                if len(set_charts_versions) > 1 or len(set_service_versions) > 1:
                    raise Exception('oops, something went wrong during previous deployment. Required human review')

                if not self.assembly_target.recreate:
                    if component_config.tag in set_service_versions and component_config.helm_chart_version in set_charts_versions:
                        write_stdout(f'Service {release_name} with version {component_config.tag} already released')
                        write_stdout(f'Service {release_name} resources were crated by chart version {component_config.helm_chart_version}')
                        continue

                # get current release version, to rollback in case of
                # every success deployment creates new release from helm perspective,
                # even if nothing has changed (because of migration command)
                last_version = run_command(
                    [
                        f"helm history {release_name} "
                        f"-n {release_name} "
                        f"--kubeconfig {kubeconfig_file_path} "
                        "| tail -n1 "
                        "| awk '{print $1}' ",
                    ],
                    cwd=component_directory,
                ).stdout

                # some low-resource kbnodes have problems with correct deploy process. Its probably easier to stop all
                # pods, and then recreate them.
                if component_config.force_disable:
                    for deployment in list_deployments:
                        kubernetes_client.patch_namespaced_deployment_scale(
                            name=deployment.metadata.name,
                            namespace=deployment.metadata.namespace,
                            body={'spec': {'replicas': 0}},
                        )
                    time.sleep(KUEBERNETES_SLEEP_TIME)

                # download helm chart
                run_command(
                    [
                        f'helm pull {component_config.helm_repo_name}/{component_config.helm_chart_name} --untar --version {component_config.helm_chart_version}',
                    ],
                    cwd=component_directory,
                )
                if not os.path.exists(os.path.join(cwd, component_config.directory, '.helm/templates')):
                    shutil.copytree(
                        os.path.join(cwd, component_config.directory, f"{component_config.helm_chart_name}/templates"),
                        os.path.join(cwd, component_config.directory, '.helm/templates'),
                    )
                    shutil.copy(  # We copy this to have more detail info in helm history
                        os.path.join(cwd, component_config.directory, f"{component_config.helm_chart_name}/Chart.yaml"),
                        os.path.join(cwd, component_config.directory, '.helm/Chart.yaml'),
                    )
                shutil.copy(os.path.join(values_dir, "secret-values.yaml"), component_directory)
                shutil.copy(os.path.join(values_dir, "values.yaml"), component_directory)
                shutil.copy(os.path.join(cwd, component_config.directory, '.helm/charts.yaml'), component_directory)
                enc_key = os.environ[self.assembly_target.encryption_key_name]
                deploy_result = run_command(
                    [
                        f'git add . '
                        f'&& export RELEASE_TAG={component_config.tag} '
                        f'&& export WERF_SECRET_KEY={enc_key} '
                        f'&& export WERF_ENV={self.assembly_target.environment} '
                        f'&& werf converge --config=werf-release.yaml --kube-config {kubeconfig_file_path} --namespace={release_name} --values=values.yaml  --values=charts.yaml --set=component_version={component_config.tag} --secret-values=secret-values.yaml --auto-rollback=true --loose-giterminism=True --dev=True'  # noqa
                    ],
                    cwd=component_directory,
                    ignore_errors=True,
                )
                if deploy_result.code != 0:
                    for release, version in releases.items():
                        # rollback all successful releases
                        run_command([
                            f'werf helm rollback {release} -n {release} {version} --kube-config {kubeconfig_file_path}',
                        ])
                    raise Exception(f'Deploy of {release_name} version {component_config.tag} failed due to some deploy problems. All deployed services were rollbacked')
                releases[release_name] = str(last_version).strip('\n')  # because run_command adds \n and i have no idea hpw to disable it



import argparse
import configparser
import json
import os
import platform
import re
import subprocess
import sys
from typing import List, Any, Tuple, Optional, Dict, Pattern, Set

import tomli
import yaml

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.commands.conf import MAX_LINE_LEN, THIS_LIB_PATH, PYENV_PY_VERSION, SUPPORTED_PYTHON_VERSIONS, MYPY_CONFIG, PIPFILE, PRE_PUSH_DEST, \
    PRE_COMMIT_DEST, UNDER_CI_JOB, PY_TYPED, SOURCE_PYPI, SOURCE_NERO_GITLAB, SOURCE_NERO_GITLAB_HOST, SOURCE_NERO_GITLAB_URL, PIPFILE_LOCK
from ul_py_tool.utils.arg_files_glob import arg_files_glob, arg_file_glob_compile_files, arg_files_print, arg_file_info
from ul_py_tool.utils.arg_str2bool import arg_str2bool
from ul_py_tool.utils.colors import FG_RED, NC, FG_YELLOW, FG_GRAY
from ul_py_tool.utils.pipfile import Pipfile
from ul_py_tool.utils.step import StepError, Stepper, Step
from ul_py_tool.utils.write_stdout import write_stdout


class CmdLint(Cmd):
    setup_file_lists: List[List[str]]
    fix: bool
    py_file_lists: List[List[str]]
    py_file_ignore_lists: List[List[str]]
    # docker_registry_tpl: str
    dockerfiles_lists: List[List[str]]
    dockerfiles_ignore_lists: List[List[str]]
    dockercomposefiles_lists: List[List[str]]
    dockercomposefiles_ignore_lists: List[List[str]]
    helmtemplates_lists: List[List[str]]
    helmtemplates_ignore_lists: List[List[str]]
    yml_file_lists: List[List[str]]
    yml_file_ignored_lists: List[List[str]]
    print_files: int
    check_environment: bool
    check_imports: bool
    exclude_libs: List[str]
    packages_ignored: List[str]
    py_init_dir_ignore_lists: List[List[str]]
    packages_ignored_in_setup: List[str]
    check_hooks: bool
    check_setup_packages: bool
    check_nero_packages_is_fresh: bool
    check_packages: bool

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--fix', dest='fix', type=arg_str2bool, default=False)
        parser.add_argument('--check-env', dest='check_environment', type=arg_str2bool, default=False)
        parser.add_argument('--check-hooks', dest='check_hooks', type=arg_str2bool, required=False, default=not UNDER_CI_JOB)
        parser.add_argument('--check-imports', dest='check_imports', type=arg_str2bool, default=True, required=False)
        parser.add_argument('--check-packages', dest='check_packages', type=arg_str2bool, default=not UNDER_CI_JOB, required=False)
        parser.add_argument('--check-setup-packages', dest='check_setup_packages', type=arg_str2bool, default=not UNDER_CI_JOB, required=False)
        parser.add_argument('--check-nero-packages-is-fresh', dest='check_nero_packages_is_fresh', type=arg_str2bool, default=not UNDER_CI_JOB, required=False)
        parser.add_argument('--ignore-setup-packages', dest='packages_ignored_in_setup', type=str, default=[], nargs='+', required=False)
        parser.add_argument('--ignore-packages', dest='packages_ignored', type=str, default=[], nargs='+', required=False)

        parser.add_argument('--exclude-import', dest='exclude_libs', type=str, default=[], nargs='+', required=False)

        parser.add_argument('--print-files', dest='print_files', type=int, required=False, default=10)

        #  parser.add_argument('--docker-registry-tpl', dest='docker_registry_tpl', type=str, required=False, default='gitlab.neroelectronics.by:5050/unic-lab')

        parser.add_argument('--dockerfiles-ignore', dest='dockerfiles_ignore_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[])
        parser.add_argument('--dockerfiles', dest='dockerfiles_lists', nargs='+', type=arg_files_glob(), required=False, default=[
            arg_files_glob(ignore_absent=True)('Dockerfile'),
            arg_files_glob(ignore_absent=True)('**/Dockerfile'),
        ])

        parser.add_argument('--dockercomposefiles-ignore', dest='dockercomposefiles_ignore_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[])
        parser.add_argument('--dockercomposefiles', dest='dockercomposefiles_lists', nargs='+', type=arg_files_glob(), required=False, default=[
            arg_files_glob(ignore_absent=True)('**/*.docker-compose.yml'),
            arg_files_glob(ignore_absent=True)('**/docker-compose.*.yml'),
            arg_files_glob(ignore_absent=True)('**/docker-compose.yml'),
            arg_files_glob(ignore_absent=True)('*.docker-compose.yml'),
            arg_files_glob(ignore_absent=True)('docker-compose.*.yml'),
            arg_files_glob(ignore_absent=True)('docker-compose.yml'),
        ])

        parser.add_argument('--helmtemplatefiles-ignore', dest='helmtemplates_ignore_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[])
        parser.add_argument('--helmtemplatefiles', dest='helmtemplates_lists', nargs='+', type=arg_files_glob(), required=False, default=[
            arg_files_glob(ignore_absent=True)('.helm/templates/**/*.yml'),
            arg_files_glob(ignore_absent=True)('.helm/templates/*.yml'),
        ])

        parser.add_argument('--setup-files', dest='setup_file_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[
            arg_files_glob(ignore_absent=True)('setup.py'),
        ])

        parser.add_argument('--py-init-dir-ignore', dest='py_init_dir_ignore_lists', nargs='+', type=arg_files_glob(ignore_absent=True, dir=True), required=False, default=[
            arg_files_glob(ignore_absent=True, dir=True)('__init__.py'),
        ])

        parser.add_argument('--py-files-ignore', dest='py_file_ignore_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[
            arg_files_glob(ignore_absent=True)('**/.*/**/*.py'),
        ])
        parser.add_argument('--py-files', dest='py_file_lists', nargs='+', type=arg_files_glob(), required=False, default=[
            arg_files_glob(ignore_absent=True)('**/*.py'),
        ])

        parser.add_argument('--yml-files-ignore', dest='yml_file_ignored_lists', nargs='+', type=arg_files_glob(ignore_absent=True), required=False, default=[
        ])
        parser.add_argument('--yml-files', dest='yml_file_lists', nargs='+', type=arg_files_glob(), required=False, default=[
            arg_files_glob(ignore_absent=True)('*.yml'),
            arg_files_glob(ignore_absent=True)('**/*.yml'),
        ])

    def _check_type_ignore_file(self, files: List[str]) -> None:
        ingore_re = re.compile(r'^\s*#\s*type\s*:\s*ignore\s*')
        noqa_re = re.compile(r'^\s*#\s*noqa\s*$')
        flake8_noqa_re = re.compile(r'^\s*#\s*flake8\s*:\s*noqa\s*$')
        type_ignored_files = []
        noqa_files = []
        for file in files:
            with open(file, 'rt') as f:
                for line in f.readlines():
                    if ingore_re.match(line) is not None:
                        type_ignored_files.append(file)
                        break
                    if noqa_re.match(line) is not None:
                        noqa_files.append(file)
                        break
                    if flake8_noqa_re.match(line) is not None:
                        noqa_files.append(file)
                        break
        if len(type_ignored_files) > 0 or len(noqa_files) > 0:
            for ign_file in type_ignored_files:
                write_stdout(f'{arg_file_info(ign_file)} :: has "# type: ignore" for HOLE FILE. please explicitly ignore this file from linting recipe')
            for ign_file in noqa_files:
                write_stdout(f'{arg_file_info(ign_file)} :: has "# noqa" or "# flake8: noqa" for HOLE FILE. please explicitly ignore this file from linting recipe')
            raise StepError()

    def _check_isort(self, result: 'subprocess.CompletedProcess[bytes]') -> None:
        lines_count = 0
        for line in result.stderr.decode('utf-8').split('\n'):
            line = line.strip()
            if not line:
                continue
            lines_count += 1
            write_stdout(line.replace('ERROR: ', '').replace(' Imports ', f':0 {FG_RED}Imports ') + NC)
        write_stdout(f'{lines_count} errors')
        if lines_count > 0:
            raise StepError()

    def _step_check_setup_files(self) -> bool:
        setup_files, ignored_files = arg_file_glob_compile_files(self.setup_file_lists)

        if len(setup_files) == 0:
            return False

        arg_files_print(self.print_files, setup_files, print_total=False, name='setup files')
        for setup_file in setup_files:
            with open(setup_file, 'rt') as f:
                setup_file_content = f.read()
                if PY_TYPED not in setup_file_content:
                    raise StepError(f'"py.typed" must be specified in {setup_file}. please add "py.typed" EMPTY file to repo and specify it in setup.py')
        return True

    def _check_imports(self, py_files: List[str]) -> bool:
        excl_libs = set(self.exclude_libs or [])
        libs = {i for i in {'flask', 'requests', 'flask_limiter', 'werkzeug'} if i not in excl_libs}

        if len(libs) == 0:
            return False

        write_stdout(f'Checking libs: {", ".join(libs)}')
        regs: Dict[str, List[Pattern[str]]] = {}
        for lib in libs:
            regs[lib] = [
                re.compile(rf'^\s*import\s+{lib}\s+'),
                re.compile(rf'^\s*import\s+{lib}\.'),
                re.compile(rf'^\s*from\s+{lib}\s+'),
                re.compile(rf'^\s*from\s+{lib}\.'),
            ]

        errors: Dict[str, List[str]] = {}
        for py_file in py_files:
            with open(py_file, 'rt') as f:
                for i, line in enumerate(f.readlines()):
                    for lib, reg_list in regs.items():
                        for reg in reg_list:
                            for m in reg.findall(line):
                                if py_file not in errors:
                                    errors[py_file] = []
                                errors[py_file].append((
                                    f'{py_file}:{i + 1}: '
                                    f'{FG_RED}RESTRICTED{NC} import of "{FG_RED}{lib}{NC}" in line #{i} '
                                    f'::    {line.replace(m, f"{FG_YELLOW}{m.strip()}{NC}").strip()}'
                                ))
        if len(errors.keys()) > 0:
            for _1, err_list in errors.items():
                for err in err_list:
                    write_stdout(err)
            raise StepError()
        return True

    def _step_check_gitignore(self) -> bool:
        gitignore = os.path.join(os.getcwd(), '.gitignore')
        if not os.path.isfile(gitignore):
            write_stdout(f'file "{gitignore}" was not found ')
            raise StepError()
        with open(gitignore, 'rt') as f:
            for line in f.readlines():
                if PYENV_PY_VERSION in line:
                    raise StepError(f'found {PYENV_PY_VERSION} in version. REMOVE IT FROM .gitignore')
        return True

    def _step_check_python_version(self) -> None:
        python_version_fp = os.path.join(os.getcwd(), PYENV_PY_VERSION)
        if not os.path.exists(python_version_fp):
            raise StepError(f'file ./{PYENV_PY_VERSION}:0 not found. please use pyenv!')
        with open(python_version_fp, 'rt') as f:
            version_string = f.read().strip()
        m = re.compile(r'^\d+\.\d+\.\d+$').findall(version_string)
        if len(m) != 1:
            raise StepError(f'invalid format of python version. {version_string} was given')

        version: str = m[0]
        minor_version = ".".join(version.split('.')[:2])
        if not any(version.startswith(v) for v in SUPPORTED_PYTHON_VERSIONS):
            raise StepError(f'unsupported python version. MUST BE {" or ".join(SUPPORTED_PYTHON_VERSIONS)}. {version} was given')
        if not version.startswith('.'.join(platform.python_version_tuple()[:2])):
            write_stdout(f'unsupported version of python to run this script. must be {minor_version}. {sys.version} was ran')

        mypy_config_fp = os.path.join(os.getcwd(), MYPY_CONFIG)
        if os.path.exists(mypy_config_fp):
            config = configparser.ConfigParser()
            config.read(mypy_config_fp)
            mypy_section = config['mypy']
            if not mypy_section:
                raise StepError(f'[mypy] section must be specified in {mypy_config_fp}')
            mypy_py_version = mypy_section.get('python_version', None)
            if not mypy_py_version:
                raise StepError(f'python_version must be specified in {mypy_config_fp}')
            if not mypy_py_version.startswith(minor_version):
                raise StepError(f'python_version in {MYPY_CONFIG} is not compatible with version in {PYENV_PY_VERSION}. must be {minor_version}')

        pipfile = os.path.join(os.getcwd(), PIPFILE)
        if not os.path.exists(pipfile):
            raise StepError(f'file {pipfile} must be specified')

        with open(pipfile, 'rt') as f:
            pipfile_content = tomli.loads(f.read())

        pipfile_requires = pipfile_content.get('requires', None)
        if pipfile_requires is None:
            raise StepError(f'[requires] section must be specified in {PIPFILE}')

        pipfile_py_version: Optional[str] = pipfile_requires.get('python_version', None)
        if pipfile_py_version is None:
            raise StepError(f'python_version must be specified in {PIPFILE}')

        if not pipfile_py_version.startswith(minor_version):
            raise StepError(f'python_version in {PIPFILE} not compatible with version in {PYENV_PY_VERSION}. must be {minor_version}')

    def _step_check_hooks(self) -> bool:
        if not self.check_hooks:
            return False

        with open(PRE_COMMIT_DEST, 'rt') as f:
            content = f.read()
            if 'pipenv run lint' not in content:
                raise StepError('"pipenv run lint" must be in pre-commit hook. please do "ulpytool install"')

        with open(PRE_PUSH_DEST, 'rt') as f:
            content = f.read()
            if 'pipenv run test' not in content:
                raise StepError('"pipenv run test" must be in pre-push hook. please do "ulpytool install"')
            if 'pipenv run lint' not in content:
                raise StepError('"pipenv run lint" must be in pre-push hook. please do "ulpytool install"')
        return True

    def _check_docker_image(self, file: str, stp: Step, image: Optional[str], *, line: int = 0, err_prefix: str = '') -> None:
        if image is None:
            return
        if len(image) == 0:
            return
        if 'WERF_' in image:
            return
        # if self.docker_registry_tpl not in image:
            # stp.add_error_in_file(file, f'{err_prefix} invalid image. must starts from "{self.docker_registry_tpl}". "{image}" was given', line=line)

    def _step_check_docker_compose_files(self, stp: Step) -> bool:
        dockercomposefiles, ignored_files = arg_file_glob_compile_files(self.dockercomposefiles_lists, self.dockercomposefiles_ignore_lists)
        if len(dockercomposefiles) == 0:
            return False
        arg_files_print(self.print_files, dockercomposefiles, ignored_files=ignored_files, name='docker-compose files')
        for dcf in dockercomposefiles:
            with open(dcf, 'rt') as f:
                content = yaml.load(f.read(), yaml.SafeLoader)
            if content.get('version') != '3.8':
                stp.add_error_in_file(dcf, 'invalid specification version. must be 3.8')
            if 'services' not in content:
                stp.add_error_in_file(dcf, 'no services found')
                continue
            for service_name, sv in content['services'].items():
                self._check_docker_image(dcf, stp, sv.get('image', None), err_prefix=f'service "{FG_RED}{service_name}{NC}" has ')
        return True

    def _step_check_dockerfiles(self, stp: Step) -> bool:
        dockerfiles, ignored_files = arg_file_glob_compile_files(self.dockerfiles_lists, self.dockerfiles_ignore_lists)
        if len(dockerfiles) == 0:
            return False
        arg_files_print(self.print_files, dockerfiles, ignored_files=ignored_files, name='Dockerfiles')
        for df in dockerfiles:
            with open(df, 'rt') as f:
                df_lines = f.readlines()
            for li, lc in enumerate(df_lines):
                lc = lc.strip()
                if lc.startswith('FROM'):
                    res = re.split(r'\s+', lc)
                    self._check_docker_image(df, stp, res[1], line=li + 1)
                if lc.startswith('RUN'):
                    if 'pipenv' in lc and 'install' in lc:
                        valid_pipenv_install_str = 'RUN pipenv install --dev --system --deploy --ignore-pipfile'
                        if lc != valid_pipenv_install_str:
                            stp.add_error_in_file(df, f'invalid command of packages installation from Pipfile. must be "{valid_pipenv_install_str}". "{lc}" was given', line=li + 1)
        return True

    def _find_helm_spec(self, spec: Any, prop: str, *, _context: str = '') -> List[Tuple[Any, str]]:
        res: List[Tuple[Any, str]] = []
        if isinstance(spec, dict):
            if prop in spec:
                res.append((spec[prop], _context[1:] if _context.startswith('.') else _context))
            for k, v in spec.items():
                res = [*res, *self._find_helm_spec(v, prop, _context=f'{_context}.{k}')]
        if isinstance(spec, list):
            for i, v in enumerate(spec):
                res = [*res, *self._find_helm_spec(v, prop, _context=f'{_context}.{i}')]
        return res

    def check_helm(self, stp: Step) -> bool:
        helm_files, ignored_files = arg_file_glob_compile_files(self.helmtemplates_lists, self.helmtemplates_ignore_lists)
        if len(helm_files) == 0:
            return False
        for f_name in helm_files:
            with open(f_name, 'rt') as f:
                contents = ['']
                for line in f.read().split('\n'):
                    if line.startswith('#'):
                        continue
                    line = re.sub(r'\{\{[^}]+\}\}', '', line)
                    if len(line.strip()) == 0:
                        continue
                    if line == '---':
                        if len(contents[-1]) > 0:
                            contents.append('')
                        continue
                    contents[-1] += f'\n{line}'
                yml_contents = [yaml.load(content, yaml.SafeLoader) for content in contents]
                for yml_content in yml_contents:
                    for image in self._find_helm_spec(yml_content, 'image'):
                        self._check_docker_image(f_name, stp, image[0], err_prefix=f'{image[1]} :: ')
        return True

    def _step_check_mypy_config(self) -> bool:
        mypy_config = os.path.join(os.getcwd(), MYPY_CONFIG)
        if not os.path.exists(mypy_config):
            return False

        config = configparser.ConfigParser()
        config.read(mypy_config)
        mypy_section = config['mypy']
        if not mypy_section:
            raise StepError(f'[mypy] section must be specified in {mypy_config}')

        # important_plugins = ['pydantic.mypy']
        # for plug in important_plugins:
        #     if plug not in mypy_section.get('plugins', '').lower():
        #         raise StepError(f'{arg_file_info(MYPY_CONFIG)} :: plugins must contains "{plug}"')
        if mypy_section.get('ignore_missing_imports', '').lower() != "false":
            raise StepError(f'{arg_file_info(MYPY_CONFIG)} :: ignore_missing_imports must be False. "{mypy_section["ignore_missing_imports"]}" was given')
        if mypy_section.get('follow_imports', '').lower() != "error":
            raise StepError(f'{arg_file_info(MYPY_CONFIG)} :: follow_imports must be error. "{mypy_section["follow_imports"]}" was given')
        if mypy_section.get('follow_imports_for_stubs', '').lower() != "true":
            raise StepError(f'{arg_file_info(MYPY_CONFIG)} :: follow_imports_for_stubs must be True. "{mypy_section["follow_imports_for_stubs"]}" was given')

        return True

    @property
    def _ignored_packages(self) -> Set[str]:
        res = {'wheel', 'setuptools'}
        for p in self.packages_ignored:
            res.add(p.strip().lower())
        return res

    @property
    def _packages_ignored_in_setup(self) -> Set[str]:
        res = set()
        for p in self.packages_ignored_in_setup:
            res.add(p.strip().lower())
        return res

    def _step_check_pipfile(self, stp: Step) -> bool:
        err_pref = f'{arg_file_info(PIPFILE, line=1)} ::'
        pipfile = os.path.join(os.getcwd(), PIPFILE)
        pipfile_lock = os.path.join(os.getcwd(), PIPFILE_LOCK)

        if not os.path.exists(pipfile):
            raise StepError(f'{err_pref} must be specified')

        if not os.path.exists(pipfile_lock):
            raise StepError(f'{err_pref} was not locked. please make "pipenv install"')

        try:
            with open(pipfile_lock, 'rt') as f:
                pipfile_lock_cntnt: Dict[str, Any] = json.load(f)
            ppfile = Pipfile.load(pipfile)
        except Exception as e:  # noqa: B902
            raise StepError(f'{err_pref} {e}')

        if len(ppfile.sources) == 0:
            raise StepError(f'{err_pref} has no sources')

        # valid_pypi_source = 'https://pypi.org/simple'
        # if ppfile.default_source.url != valid_pypi_source:
        #     stp.add_error_in_file(pipfile, f'has invalid default source url. must be "{valid_pypi_source}". {ppfile.default_source.url} was given')

        if not ppfile.default_source.verify_ssl:
            stp.add_error_in_file(pipfile, f'default source must be verified with ssl. please set "verify_ssl = true" for source "{ppfile.default_source.name}"')

        if ppfile.default_source.name != SOURCE_PYPI:
            stp.add_error_in_file(pipfile, f'default source must have name "{SOURCE_PYPI}". "{ppfile.default_source.name}" was given')

        for source in ppfile.sources:
            if SOURCE_NERO_GITLAB_HOST not in source.url:
                continue
            if source.name != SOURCE_NERO_GITLAB:
                stp.add_error_in_file(pipfile, f'please rename source "{FG_RED}{source.name}{NC}" to {SOURCE_NERO_GITLAB}')
            if source.url != SOURCE_NERO_GITLAB_URL:
                stp.add_error_in_file(pipfile, f'ALL GITLAB packages must be stored in one project. url must be "{SOURCE_NERO_GITLAB_URL}"')
            if source.verify_ssl:
                stp.add_error_in_file(pipfile, f'source "{source.name}" must have "verify_ssl = false"')

        if self.check_hooks:
            for command in {'test', 'lint'}:
                if ppfile.scripts.get(command, None) is None:
                    stp.add_error_in_file(pipfile, f'script "{FG_RED}{command}{NC}" is not specified')
                else:
                    if not ppfile.scripts[command].command:
                        stp.add_error_in_file(pipfile, f'script "{FG_RED}{command}{NC}" has empty command')

        setup_files, _1 = arg_file_glob_compile_files(self.setup_file_lists)

        main_setup_file_content = ''
        for sf in setup_files:
            if os.path.dirname(sf) == os.getcwd():
                with open(sf, 'rt') as f:
                    main_setup_file_content = f.read()
                break

        installed_packages = Pipfile.freeze()
        for pack in ppfile.all_packages.values():
            if pack.name in self._ignored_packages or not self.check_packages:
                continue

            assert pack.source is not None

            if pack.semver is None and pack.source.name == SOURCE_NERO_GITLAB:
                stp.add_error_in_file(
                    pipfile,
                    f'package "{FG_RED}{pack.name}{NC}" form "{SOURCE_NERO_GITLAB}" must have version with valid semver formant. '
                    f'IT IS THE RULE OF COMPANY. {pack.version} was given',
                )

            if pack.version_operator != '==':
                stp.add_error_in_file(
                    pipfile,
                    f'package "{FG_RED}{pack.name}{NC}" has invalid operator for version. must be "==". '
                    f'"{pack.version_operator}" was given in version "{pack.version}"',
                )

            if pack.source.name == SOURCE_NERO_GITLAB:
                if pack.is_dev:
                    stp.add_error_in_file(pipfile, f'package "{FG_RED}{pack.name}{NC}" could not be in the dev-package!')

                if len(pack.available_versions) > 0:
                    msg = f'package "{FG_RED}{pack.name}{NC}" is not fresh. version={FG_RED}{pack.semver}{NC}. ' \
                        f'versions for upgrading: {", ".join(str(v) for v in pack.available_versions)}'

                    if pack.name == 'ulpytool':
                        stp.add_error(msg)
                    else:
                        stp.add_warn(msg)

            if pack.name not in installed_packages:
                stp.add_error_in_file(pipfile, f'package "{FG_RED}{pack.name}{NC}" not installed')
            else:
                if installed_packages[pack.name].clean_version != pack.clean_version:
                    stp.add_error_in_file(
                        pipfile,
                        f'package "{FG_RED}{pack.name}{NC}" installed version is invalid. '
                        f'must be "{pack.clean_version}". '
                        f'{installed_packages[pack.name].clean_version} was given',
                    )

            if pack.is_dev:
                ps = pipfile_lock_cntnt['develop']
            else:
                ps = pipfile_lock_cntnt['default']
            if pack.name not in ps:
                stp.add_error_in_file(pipfile, f'package "{FG_RED}{pack.name}{NC}" not in lock. please make "pipenv install{" --dev" if pack.is_dev else ""}"')
            else:
                p = ps[pack.name]
                if p['version'] != pack.version:
                    stp.add_error_in_file(
                        pipfile,
                        f'package "{FG_RED}{pack.name}{NC}" has invalid version. '
                        f'must be {FG_RED}"{pack.version}"{NC}. '
                        f'"{p["version"]}" was given',
                    )

            if main_setup_file_content and pack.name not in self._packages_ignored_in_setup and not pack.is_dev and pack.source.name != SOURCE_NERO_GITLAB:
                if pack.pip_representation not in main_setup_file_content and pack.pip_representation.replace('==', '>=') not in main_setup_file_content:
                    stp.add_error_in_file(pipfile, f'package "{FG_RED}{pack.pip_representation}{NC}" not in {arg_file_info(setup_files[0])}')
        return True

    def _chk_ymllinter(self, result: 'subprocess.CompletedProcess[bytes]') -> None:
        form_re = re.compile(r'^(.+?):(\d+):(\d+):\s+\[([^]]+)]\s+([^(]+)\(([^)]+)\)\s*$')
        errors: List[Tuple[str, str, str, str, str, str]] = []
        max_len_of_file = 0
        for lc in result.stdout.decode().split('\n'):
            m = form_re.findall(lc.strip())
            for mr in m:
                if len(mr) == 6:
                    file, line, col, lvl, msg, code = mr
                    max_len_of_file = max(max_len_of_file, len(file))
                    errors.append(mr)
        for file, line, col, lvl, msg, code in errors:
            write_stdout(f'{arg_file_info(os.path.abspath(file), line=int(line), col=int(col)): <{max_len_of_file + 10}} :: {FG_RED}{lvl}{NC} :: {msg} :: {FG_GRAY}({code}){NC}')
        if len(errors) == 0 and result.returncode != 0:
            write_stdout(result.stdout.decode())
        result.check_returncode()

    def _check_init_files(self, py_files: List[str], stp: Step) -> None:
        uniq_init_files = set()
        cwd = os.path.abspath(os.getcwd())
        for f in py_files:
            d = os.path.dirname(f)
            if d != cwd:
                uniq_init_files.add(os.path.join(d, '__init__.py'))

        init_files, ignored_inits = arg_file_glob_compile_files([list(uniq_init_files)], self.py_init_dir_ignore_lists)
        for init in init_files:
            if not os.path.exists(init):
                stp.add_error_in_file(init, 'is not exists')
            elif not os.path.isfile(init):
                stp.add_error_in_file(init, 'is not a file')

        for init in init_files:
            if not os.path.exists(init):
                continue
            with open(init, 'rt') as init_f:
                if len(init_f.read().strip()) > 0:
                    stp.add_warn(f'init file "{init}" is not empty')

    def run(self) -> None:
        stepper = Stepper()

        with stepper.step('CHECK PYTHON VERSION'):
            self._step_check_python_version()

        has_py_files = self._check_py_files(stepper)

        with stepper.step('CHECK HOOKS', print_error=True) as stp:
            if not self._step_check_hooks():
                stp.skip()

        yml_files, ignored_yml_files = arg_file_glob_compile_files(self.yml_file_lists, self.yml_file_ignored_lists)
        with stepper.step('CHECK YAML FILES') as stp:
            if len(yml_files) == 0:
                stp.skip()
            else:
                arg_files_print(self.print_files, yml_files, ignored_files=ignored_yml_files)
                stp.run_cmd([
                    'yamllint',
                    '--config-file', os.path.join(THIS_LIB_PATH, 'conf', 'yamlint.yml'),
                    '--format', 'parsable',
                    '--strict',
                    *yml_files,
                ], chk=self._chk_ymllinter, print_std=False)

        with stepper.step('CHECK HELM-TEMPLATES') as stp:
            if not self.check_helm(stp):
                stp.skip()

        with stepper.step('CHECK PIPFILE') as stp:
            if not self._step_check_pipfile(stp):
                stp.skip()

        with stepper.step('CHECK DOCKER COMPOSE FILES') as stp:
            if not self._step_check_docker_compose_files(stp):
                stp.skip()

        with stepper.step('CHECK SETUP FILES') as stp:
            if not self._step_check_setup_files():
                stp.skip()

        with stepper.step('CHECK gitignore') as stp:
            if not self._step_check_gitignore():
                stp.skip()

        with stepper.step('CHECK DOCKERFILES') as stp:
            if not self._step_check_dockerfiles(stp):
                stp.skip()

        if not has_py_files:
            raise StepError('no python files found for linting')

    def _check_py_files(self, stepper: Stepper) -> bool:
        py_files, ignored_files = arg_file_glob_compile_files(self.py_file_lists, self.py_file_ignore_lists)
        write_stdout('')
        arg_files_print(self.print_files, py_files, ignored_files=ignored_files, name='python files')

        local_mypy_conf = os.path.join(os.getcwd(), 'mypy.ini')
        mypy_conf = local_mypy_conf if os.path.isfile(local_mypy_conf) else os.path.join(THIS_LIB_PATH, "conf", "mypy.ini")

        with stepper.step('CHECK __init__.py FILES') as stp:
            if not len(py_files):
                stp.skip()
            else:
                self._check_init_files(py_files, stp)

        with stepper.step('TYPE IGNORE FILES') as stp:
            if not len(py_files):
                stp.skip()
            else:
                self._check_type_ignore_file(py_files)

        with stepper.step('CHECK MYPY CONFIG') as stp:
            if not len(py_files):
                stp.skip()
            else:
                if not self._step_check_mypy_config():
                    stp.skip()

        with stepper.step('MYPY') as stp:
            if not len(py_files):
                stp.skip()
            else:
                stp.run_cmd([
                    'mypy',
                    '--follow-imports=normal',
                    f'--config-file={mypy_conf}',
                    '--namespace-packages',
                    '--explicit-package-bases',
                    *py_files,
                ])

        with stepper.step('LINTING RUFF') as stp:
            if not len(py_files):
                stp.skip()
            else:
                stp.run_cmd([
                    'ruff',
                    'check',
                    *(('--fix', ) if self.fix else tuple()),
                    '--cache-dir',
                    os.path.join(os.getcwd(), ".tmp", "ruff_cache"),
                    f'--line-length={MAX_LINE_LEN}',
                    *py_files,
                ])

        with stepper.step('LINT IMPORTS') as stp:
            if not len(py_files):
                stp.skip()
            else:
                if not self.check_imports:
                    stp.skip()
                elif not self._check_imports(py_files):
                    stp.skip()

        return bool(py_files)



import argparse
from typing import Union


def arg_str2bool(value: Union[str, int, bool]) -> bool:
    if isinstance(value, str):
        return value.lower() not in {'0', 'false', '', 'no'}

    if isinstance(value, int):
        return bool(value)

    if isinstance(value, bool):  # type: ignore
        return value

    raise argparse.ArgumentTypeError('invalid type')



import re
from typing import NamedTuple, List, Dict, Optional

from ul_py_tool.utils.run_command import CommandResult, run_command

FROM_RE = re.compile(r'^FROM\s+(?P<image>[^\s]+)(?:\s+as\s+(?P<label>[^\s]+))?$')


class DockerFile(NamedTuple):
    file_path: Optional[str]
    lines: List[str]
    images: Dict[str, str]  # label => image with version

    def ensure_image(self, label: str = '') -> str:
        return self.images[label]

    @staticmethod
    def load_image(image: str) -> 'DockerFile':
        assert isinstance(image, str)
        return DockerFile(
            file_path=None,
            lines=[],
            images={
                '': image,
            },
        )

    def push(self, full_url: str, silent: bool = False) -> CommandResult:
        return run_command([f'docker push {full_url}'], silent=silent)

    def build(self, full_url: str, build_args: List[str], build_dir: str = '.', silent: bool = False) -> CommandResult:
        build_args_str = f"--build-arg {' --build-arg '.join(build_args)}" if len(build_args) else ""
        return run_command([f"docker build -f {self.file_path} -t {full_url} {build_dir} {build_args_str}"], silent=silent)

    @staticmethod
    def load_file(docker_file_path: str, image_name: str = '') -> 'DockerFile':
        images = dict()
        images[''] = image_name
        lines: List[str] = []

        with open(docker_file_path, 'rt') as dockery:
            for line in dockery:
                if FROM_RE.match(line):
                    groups_r = FROM_RE.search(line)
                    assert groups_r is not None
                    groups = groups_r.groupdict()
                    images[groups.get('label', '')] = groups['image']

        return DockerFile(
            file_path=docker_file_path,
            lines=lines,
            images=images,
        )



import argparse
import os.path
import re
from datetime import datetime
from enum import Enum
from typing import Tuple

from ul_py_tool.commands.cmd import Cmd
from ul_py_tool.commands.conf import CHANGE_LOG
from ul_py_tool.utils.arg_file_exists import arg_file_exists
from ul_py_tool.utils.colors import FG_GREEN, NC, FG_BLUE
from ul_py_tool.utils.input_lines import input_lines
from ul_py_tool.utils.step import Stepper, StepError
from ul_py_tool.utils.write_stdout import write_stdout


class CmdVersionType(Enum):
    PATCH = 'patch'
    MINOR = 'minor'
    MAJOR = 'major'


class CmdVersion(Cmd):
    setup_file: str
    type: CmdVersionType

    @staticmethod
    def add_parser_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument('--setup-file', dest='setup_file', type=arg_file_exists, default=os.path.join(os.getcwd(), 'setup.py'), required=False)

    @property
    def change_log_file(self) -> str:
        return os.path.join(os.path.dirname(self.setup_file), CHANGE_LOG)

    def _ask_change_logs(self, new_version: str, dt_now: datetime) -> bool:
        bugfixes_lines = input_lines(f'{FG_BLUE}bug fixes{NC}')
        new_features_lines = input_lines(f'{FG_BLUE}new features{NC}')
        release_notes_lines = input_lines(f'{FG_BLUE}release notes{NC}')

        if not len(bugfixes_lines) and not len(new_features_lines) and not len(release_notes_lines):
            return False

        logs_main_header = '# Changelog\n'

        logs_prev_content = ''
        if os.path.exists(self.change_log_file):
            with open(self.change_log_file, 'rt') as f:
                logs_prev_content = f.read()
        if not logs_prev_content:
            logs_prev_content = logs_main_header

        version_marker = '## Version `'

        log_version = f'{version_marker}{new_version}` - {dt_now.date().isoformat()}\n\n'

        for line in release_notes_lines:
            log_version += f'{line}\n\n'

        log_version += '' if not new_features_lines else '### New Features\n\n'
        for line in new_features_lines:
            log_version += f'- {line}\n'

        log_version += '\n' if new_features_lines else ''
        log_version += '' if not bugfixes_lines else '### Bug Fixes\n\n'
        for line in bugfixes_lines:
            log_version += f'- {line}\n'

        log_version += '\n'

        with open(self.change_log_file, 'wt+') as f:
            if version_marker in logs_prev_content:
                logs_prev_content = logs_prev_content.replace(version_marker, f'{log_version}\n\n{version_marker}', 1)
            else:
                logs_prev_content += f'\n\n{log_version}'
            f.write(logs_prev_content)

        return True

    def _increase_version(self) -> str:
        reg = re.compile(r'version\s*=\s*["\']\d+\.\d+\.\d+[\'"]')
        reg_version = re.compile(r'(version\s*=\s*["\'])(\d+\.\d+\.\d+)([\'"])')

        write_stdout(f'read "{self.setup_file}"')
        with open(self.setup_file, 'rt') as f:
            content = f.read()
            all_versions = reg.findall(content)
            if len(all_versions) == 0:
                write_stdout(f'semver was not found in file "{self.setup_file}"')
                raise StepError()
            if len(all_versions) > 1:
                write_stdout(f'semver was found more than once in file "{self.setup_file}"')
                raise StepError()
            match: str = all_versions[0]
            version_parts: Tuple[str, str, str] = reg_version.match(match).groups()  # type: ignore
            assert len(version_parts) == 3
            version: str = version_parts[1]
            write_stdout(f'found version {version}')

        version_segm: Tuple[int, int, int] = tuple(int(v) for v in version.split('.'))  # type: ignore

        if self.type is CmdVersionType.PATCH:
            new_version_segm = (version_segm[0], version_segm[1], version_segm[2] + 1)
            write_stdout(f'{version} -> {new_version_segm[0]}.{new_version_segm[1]}.{FG_GREEN}{new_version_segm[2]}{NC}')
        elif self.type is CmdVersionType.MINOR:
            new_version_segm = (version_segm[0], version_segm[1] + 1, 0)
            write_stdout(f'{version} -> {new_version_segm[0]}.{FG_GREEN}{new_version_segm[1]}.{new_version_segm[2]}{NC}')
        elif self.type is CmdVersionType.MAJOR:
            new_version_segm = (version_segm[0] + 1, 0, 0)
            write_stdout(f'{version} -> {FG_GREEN}{new_version_segm[0]}.{new_version_segm[1]}.{new_version_segm[2]}{NC}')
        else:
            raise NotImplementedError(f'type {self.type} was not implemented for increasing version command')
        new_version = ".".join(str(i) for i in new_version_segm)
        content = content.replace(match, f'{version_parts[0]}{new_version}{version_parts[2]}')

        with open(self.setup_file, 'wt') as f:
            f.write(content)
            write_stdout(f'Changes saved to {self.setup_file}')

        return new_version

    def run(self) -> None:
        dt_now = datetime.now()

        stepper = Stepper()

        with stepper.step('INCREASE VERSION', print_error=True):
            new_version = self._increase_version()

        with stepper.step('RELEASE NOTES') as stp:
            has_changelog_changes = self._ask_change_logs(new_version, dt_now)
            if not has_changelog_changes:
                stp.skip()

        files_to_add = [self.setup_file]
        if has_changelog_changes:
            files_to_add.append(self.change_log_file)

        with stepper.step('GIT') as stp:
            stp.run_cmd(['git', 'reset'])
            stp.run_cmd(['git', 'add', *files_to_add])
            stp.run_cmd(['git', 'commit', '-m', new_version])
            stp.run_cmd(['git', 'tag', new_version])


class CmdVersionMinor(CmdVersion):
    type: CmdVersionType = CmdVersionType.MINOR


class CmdVersionPatch(CmdVersion):
    type: CmdVersionType = CmdVersionType.PATCH


class CmdVersionMajor(CmdVersion):
    type: CmdVersionType = CmdVersionType.MAJOR
